﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Windows.Threading;
using System.Threading.Tasks; 
using Arction.Wpf.Charting;
using Arction.Wpf.Charting.Axes;
using Arction.Wpf.Charting.SeriesXY;
using Arction.Wpf.Charting.Views;
using Arction.Wpf.Charting.Views.ViewXY; 

namespace WpfApplicationBillionPoints
{
    /// <summary>
    /// LightningChart® Billion points demo. Utilizes fixed-interval data series, SampleDataSeries. 
    /// X interval = 1 / SamplingFrequency 
    /// X[i] = FirstSampleTimeStamp + (i) / SamplingFrequency. 
    /// Last data point X = FirstSampleTimeStamp + (PointsInSeries-1) / SamplingFrequency. 
    /// SampleDataSeries is configured here to use 32-bit floating point precision to save memory
    /// NOTE: The series does not down-sample data. 
    /// <remarks>Needs a quite powerful computer.
    /// Developed and tested with desktop PC: Intel i7, Nvidia GTX 960, 16 GB RAM. 
    /// Data point rate may be significantly lower with old laptops, and setting X axis range shorter is needed.   
    /// With this system, it was tested the chart can show 1.8 Billion data points before running out of memory. 
    ///</remarks>
    /// </summary>
    public partial class MainWindow : Window
    {

        LightningChartUltimate _chart;
        DispatcherTimer _timer;
        System.Diagnostics.Stopwatch _stopWatch; 
        int _seriesCount;
        int _appendCountPerRound;
        double _xLen;
        const double XInterval = 1;
        const double YMin = 0;
        const double YMax = 100;
        const int PreGenerateDataForRoundCount = 200; 
        float[][] _data;
        int _iRound = 0;
        long _pointsAppended;
        int _framesRenderedCount = 0;

        public MainWindow()
        {
            InitializeComponent();

            CreateChart();

            //Create timer 
            _timer = new DispatcherTimer();

            _timer.Interval = TimeSpan.FromMilliseconds(0);
            _timer.Tick += _timer_Tick;

            //Create StopWatch
            _stopWatch = new System.Diagnostics.Stopwatch(); 
        }

      

        void CreateChart()
        {
            _chart = new LightningChartUltimate();
            
            _chart.BeginUpdate();

            _chart.ChartRenderOptions.DeviceType = RendererDeviceType.AutoPreferD11;
            _chart.ChartRenderOptions.LineAAType2D = LineAntiAliasingType.QLAA;

            _chart.Title.Font.Size = 25;
            _chart.Title.Text = "Set options and press Start.\nPC with 16GB RAM + fast graphics card is strongly recommended";
            _chart.Title.Color = Color.FromArgb(255, 255, 204, 0);
            _chart.Title.Align = ChartTitleAlignment.TopLeft; 

            ViewXY v = _chart.ViewXY;

            //Set real-time monitoring scroll mode 
            v.XAxes[0].ScrollMode = XAxisScrollMode.Scrolling;
            v.XAxes[0].ValueType = AxisValueType.Number;
            v.XAxes[0].AutoFormatLabels = false; 
            v.XAxes[0].LabelsNumberFormat ="N0";
            v.XAxes[0].SweepingGap = 3;
            v.XAxes[0].Title.Text = "Point number"; 
            v.XAxes[0].SetRange(0, 100000); 

            //Set real-time monitoring automatic old data destruction
            v.DropOldSeriesData = true; 

            //Set Axis layout to Stacked
            v.AxisLayout.YAxesLayout = YAxesLayout.Stacked;

            v.LegendBoxes[0].Visible = false;
            
            _chart.EndUpdate(); 

            gridMain.Children.Add(_chart);
            Grid.SetRow(_chart, 0);
            Grid.SetColumn(_chart, 0);
        } 

        private void buttonStartStop_Click(object sender, RoutedEventArgs e)
        {
            //Start 

            _iRound = 0;
            _pointsAppended = 0;
            _framesRenderedCount = 0;
            
            
            CompositionTarget.Rendering -= CompositionTarget_Rendering;
            CompositionTarget.Rendering += CompositionTarget_Rendering;

            buttonStartStop.Content = "Restart";
            
            
            ViewXY v = _chart.ViewXY;

            //Read series count
            try
            {
                _seriesCount = int.Parse(textBoxSeriesCount.Text);
            }
            catch
            {
                MessageBox.Show("Invalid series count text input");
                return;
            }

            //Read append count / round
            try
            {
                _appendCountPerRound = int.Parse(textBoxAppendCountPerRound.Text);
            }
            catch
            {
                MessageBox.Show("Invalid append count");
                return;
            }

            //Read X axis length
            try
            {
                _xLen = double.Parse(textBoxXLen.Text);
            }
            catch
            {
                MessageBox.Show("Invalid X-Axis length text input");
                return;
            }

            //Create all data to be inputted before starting the monitoring, 
            //to prevent data points generating to interrupt

            _data = CreateInputData( _seriesCount, _appendCountPerRound); 


            _chart.BeginUpdate();

            CleanUpExistingObjects(v); 

            
            
            //Series count of Y axes and SampleDataSeries  
            for (int seriesIndex = 0; seriesIndex < _seriesCount; seriesIndex++)
            {
                Color lineBaseColor = DefaultColors.SeriesForBlackBackgroundWpf[seriesIndex % DefaultColors.SeriesForBlackBackgroundWpf.Length];

                AxisY axisY = new AxisY(v);
                axisY.SetRange(YMin, YMax);
                
                axisY.Title.Text = string.Format("Series {0}", seriesIndex + 1);
                axisY.Title.Angle = 90;
                axisY.Title.Color = ChartTools.CalcGradient(lineBaseColor, System.Windows.Media.Colors.White, 50);
                axisY.Units.Visible = false;
                axisY.MouseScaling = false;
                axisY.MajorGrid.Visible = false;
                axisY.MinorGrid.Visible = false; 
                v.YAxes.Add(axisY);

                SampleDataSeries series = new SampleDataSeries(v, v.XAxes[0], axisY);
                v.SampleDataSeries.Add(series);
                series.SampleFormat = SampleFormat.SingleFloat;//Use single floating point to save memory. DoubleFloat is actually faster 
                series.LineStyle.Color = ChartTools.CalcGradient(lineBaseColor, System.Windows.Media.Colors.White, 50);
                series.SamplingFrequency = 1.0 / XInterval; //Set 1 / X interval here 
                series.FirstSampleTimeStamp = 1.0 / series.SamplingFrequency;//Set first X here 
                series.LineStyle.Width = 2f;
                series.LineStyle.AntiAliasing = LineAntialias.Normal;
                series.ScrollModePointsKeepLevel = 5;
                series.ScrollingStabilizing = false;
                series.MouseInteraction = false;

                //Set custom colored palette. Color high values in red and low values in blue. 
                series.UsePalette = true; 
                series.ValueRangePalette.Steps.Clear();
                
                series.ValueRangePalette.Steps.Add(new PaletteStep(series.ValueRangePalette, ChartTools.CalcGradient(series.LineStyle.Color, System.Windows.Media.Colors.Blue, 80), 0));
                series.ValueRangePalette.Steps.Add(new PaletteStep(series.ValueRangePalette, series.LineStyle.Color, 15));
                series.ValueRangePalette.Steps.Add(new PaletteStep(series.ValueRangePalette, series.LineStyle.Color, 50));
                series.ValueRangePalette.Steps.Add(new PaletteStep(series.ValueRangePalette, series.LineStyle.Color, 85));
                series.ValueRangePalette.Steps.Add(new PaletteStep(series.ValueRangePalette, ChartTools.CalcGradient(series.LineStyle.Color, System.Windows.Media.Colors.Red, 80), 100));
                
                axisY.PaletteSeries = series; 
                axisY.UsePalette = true;
            }

            v.XAxes[0].SetRange(0, _xLen);

            //Prefill with data, this may take several seconds  
            if (checkBoxPrefill.IsChecked == true)
            {
                while (_pointsAppended < _xLen * 0.9)
                    FeedData(); 
            }
            

            _chart.EndUpdate();

            _stopWatch.Restart(); 
            _timer.Start(); 
        }


        void CleanUpExistingObjects(ViewXY v)
        {
            foreach (SampleDataSeries series in v.SampleDataSeries)
            {
                foreach (PaletteStep step in series.ValueRangePalette.Steps)
                {
                    step.Dispose(); 
                }
                series.Dispose();
            }
            v.SampleDataSeries.Clear();

            foreach (AxisY yAxis in v.YAxes)
            {
                yAxis.Dispose(); 
            }
            v.YAxes.Clear();

            GC.Collect(); 

        } 

        void CompositionTarget_Rendering(object sender, EventArgs e)
        {
            _framesRenderedCount++;
        }



        float[][] CreateInputData(int seriesCount, int appendCountPerRound)
        {
            float[][] data = new float[seriesCount][];
            System.Threading.Tasks.Parallel.For(0, seriesCount, (seriesIndex) =>
            {
                int dataPointCount = PreGenerateDataForRoundCount * appendCountPerRound;
                float[] seriesData = new float[dataPointCount];
                float seriesIndexPlus1 = (float)(seriesIndex + 1);
                Random rand = new Random((int)DateTime.Now.Ticks/(seriesIndex+1));
                
                double y = 50; 
                
                for (int i = 0; i < dataPointCount; i++)
                {
                    y = y - 0.05 + rand.NextDouble() / 10.0;
                    if (y > YMax)
                        y = YMax;
                    if (y < YMin)
                        y = YMin; 
                    seriesData[i] = (float) y;
                }
                data[seriesIndex] = seriesData; 
            });

            
            return data; 
        }


        void FeedData()
        {
            if (_chart != null)
            {
                _chart.BeginUpdate();

                //Append data
                System.Threading.Tasks.Parallel.For(0, _seriesCount, (seriesIndex) =>
                {
                    float[] thisSeriesData = _data[seriesIndex];
                    float[] dataToAppendNow = new float[_appendCountPerRound];
                    Array.Copy(thisSeriesData, _appendCountPerRound * (_iRound % PreGenerateDataForRoundCount), dataToAppendNow, 0, _appendCountPerRound);
                    _chart.ViewXY.SampleDataSeries[seriesIndex].AddSamples(dataToAppendNow, false);
                });


                _pointsAppended += _appendCountPerRound;


                //Set X axis real-time scrolling position 
                double lastX = _pointsAppended * XInterval;
                _chart.ViewXY.XAxes[0].ScrollPosition = lastX;
                
                //Update some statistics in the title
                _chart.Title.Text = string.Format(
                    "Data points in visible range: {0} | Data points total {1} | Avg FPS: {2}",
                    (Math.Min(_chart.ViewXY.XAxes[0].Maximum - _chart.ViewXY.XAxes[0].Minimum, _pointsAppended)*_seriesCount) .ToString("N0"),
                    (_pointsAppended * _seriesCount).ToString("N0"),
                    ((double)_framesRenderedCount / (double)_stopWatch.ElapsedMilliseconds * 1000.0).ToString("0.0")
                    );

                _chart.EndUpdate();

                _iRound++;

            }
        } 

         
        void _timer_Tick(object sender, EventArgs e)
        {
            FeedData();  
        }

        private void comboBoxScrollMode_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            if (_chart != null)
            {
                if (comboBoxScrollMode.SelectedIndex == 0)
                    _chart.ViewXY.XAxes[0].ScrollMode = XAxisScrollMode.Scrolling;
                else if (comboBoxScrollMode.SelectedIndex == 1)
                    _chart.ViewXY.XAxes[0].ScrollMode = XAxisScrollMode.Sweeping;
            }
        }

        private void buttonZoomXPlus_Click(object sender, RoutedEventArgs e)
        {
            ZoomX(0.5);
        }

       
        private void buttonZoomXMinus_Click(object sender, RoutedEventArgs e)
        {
            ZoomX(2);
        }

        void ZoomX(double factor)
        {
            _chart.BeginUpdate();
            AxisX xAxis = _chart.ViewXY.XAxes[0];
            double xLen = xAxis.Maximum - xAxis.Minimum;
            xAxis.SetRange(xAxis.ScrollPosition - xLen * factor, xAxis.ScrollPosition);
            _chart.EndUpdate();
        }


        private void buttonZoomYPlus_Click(object sender, RoutedEventArgs e)
        {
            ZoomY(0.5); 
        }

        private void buttonZoomYMinus_Click(object sender, RoutedEventArgs e)
        {
            ZoomY(2); 
        }

        void ZoomY(double factor)
        {
            _chart.BeginUpdate();
            foreach (AxisY yAxis in _chart.ViewXY.YAxes)
            {
                double yLen = yAxis.Maximum - yAxis.Minimum;
                double yLenNew = factor * yLen; 
                double yCenter = (yAxis.Minimum + yAxis.Maximum) / 2.0;
                yAxis.SetRange(yCenter - yLenNew / 2.0, yCenter + yLenNew / 2.0);
            }
            _chart.EndUpdate();

        }
    }
}
